/* vma2c.c -- convert .vma file to const array */
/*            or to intel hex format           */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#define MAXVOICE 128
#define NAMELEN 16
#define DATALEN 26

char instname[MAXVOICE][NAMELEN];

// linear voice array
char lva[MAXVOICE * DATALEN];

// byte copy flag array
int cbf[DATALEN];

struct ma2reg {
  unsigned char KSR : 1; // Key Scaling for Rate
  unsigned char SUS : 1; // SUStain flag
  unsigned char EGT : 1; // EG Type flag
  unsigned char VIB : 1; // VIBrato flag
  unsigned char MULT: 4; // freq MULTiplier
  unsigned char DR  : 4; // Decay   Rate
  unsigned char RR  : 4; // Release Rate
  unsigned char SL  : 4; // Sustain Level
  unsigned char AR  : 4; // Attack  Rate
  unsigned char KSL : 2; // Key Scaling for Level
  unsigned char TL  : 6; // Total   Level
  unsigned char WS  : 3; // Waveform Select
  unsigned char AM  : 1; // Amplitude Modulation flag
  unsigned char DAM : 2; // Depth of Amplitude Modulation
  unsigned char DVB : 2; // Depth of ViBrato
}; // struct ma2reg

struct voicedata {
//  unsigned char pad0;    // unused
//  unsigned char bank;    // bank LSB
//  unsigned char pcno;    // program change number
  unsigned char ALG : 3; // ALGorithm
  unsigned char FB  : 3; // FeedBack level
  unsigned char LFO : 2; // Low Frequency Osccillator
//  unsigned char pad1;    // unused
  struct ma2reg op[4];   // MA2 operator data
//  unsigned char pad2;    // unused 
}; // struct voicedata

unsigned char vbuf[DATALEN];
long int nvoice;
int hexmode;
int csum;
long int adrs;
int endrecord;
int rlen = 16;
int lva_inc = 11;
char cbf_str[DATALEN] = "00010111110";

void options(int argc, char * const argv[])
{
  int optc;
  if (1 < argc) {
    while (-1 != (optc = getopt(argc, argv, "A:a:C:c:EeFfL:l:MmS:s:VvXx"))){
      switch (optc) {
        case 'A': case 'a': // Start address
        case 'S': case 's': // Start address
          adrs = strtol(optarg, NULL, 0x10);
          break;
        case 'C': case 'c': // Copy Byte Flag
          strncpy(cbf_str, optarg, sizeof(cbf_str));
          break;
        case 'L': case 'l': // intel hex record length
          rlen = strtol(optarg, NULL, 10);
          if      (0   >= rlen) rlen = 1;
          else if (255 <  rlen) rlen = 255;
          break;
        case 'E': case 'e': // end record flag
          endrecord = 1;
          break;
        case 'X': case 'x': // intel hex file generation
        case 'V': case 'v': // for V850
          hexmode = 1;
          break;
        case 'M': case 'm': // S-format file generation
        case 'F': case 'f': // for FR60
          hexmode = 2;      
          break;
        default:
          break;
      } // switch
    } // while
  } // if (0 < argc)
} // void options()

void outhex(int x)
{
  x &= 0xff;
  printf("%.2X", x);
  csum += x;
  csum &= 0xff;
} // void outhex()

int main(int argc, char * const argv[])
{
  int i, j, k, op;
  char dum[4];
  unsigned char b;
  unsigned int bank, pc;
  struct voicedata vdat;
  char *lvap;
  char banner1[20], banner2[20];

  hexmode = 0;
  adrs = 0;
  endrecord = 0;
  options(argc, argv);
  lva_inc = 0;
  for (i = 0; i < strlen(cbf_str); i++) {
    j = ('1' == cbf_str[i]); // '1' specified?
    lva_inc += j;            // increment byte count
    if (5 <= i) {
     lva_inc += j;// increment byte count for OP2
    }
    cbf[i] = j;              // set copy flag
  } // for (i 
/*
printf("'%s' %d\r\xa", cbf_str, lva_inc);
for (i = 0; i < strlen(cbf_str); i++)
  printf("cbf[%.2d] = %d\r\xa", i, cbf[i]);
*/
  fread(dum, 4, 1, stdin);
  if (strncmp(dum, "FM  ", 4)) return(1); // check "FM  " chunk ID
  nvoice = 0;
  for (i = 0; i < 4; i++) { // get chunk length
    fread(&b, 1, 1, stdin);
    nvoice = (nvoice << 8) + b; 
  } // for (i
  nvoice /= (NAMELEN + DATALEN);
  if (MAXVOICE < nvoice) return(1); // too many data
  if (!hexmode) {
    strcpy(banner1, (cbf[0] ? "      " : ""));
    strcpy(banner2, banner1);
    for (j = 1; j < 3; j++) { //
      strcat(banner1, (cbf[j] ? "      " : ""));
    } // for (j
    strcat(banner2, (cbf[1] ? "Bank# " : ""));
    strcat(banner2, (cbf[2] ? "Inst# " : ""));
    printf("#include \"tonegen.h\"\r\xa\r\xa");
    printf("// GM voice data in MA-2 format\r\xa\r\xa");
    printf("const ma2vdat vma2[%d] PROGMEM = { //\r\xa", nvoice);
    printf("//\r\xa");
    printf("//%s FB -+-+   M  /EGT/SUS/KSR          /DAM  \r\xa", banner1);
    printf("//%s     L A   U V                 K   D      \r\xa", banner1);
    printf("//%s     F L   L I   R D   A S   T S   V A/WS    (1st line: OP1)\r\xa", banner1);
    printf("//%s     O G   T B   R R   R L   L L   B M       (2nd line: OP2)\r\xa", banner1);
    printf("//   %s  |/    |/    |/    |/    |/    |/", banner2 );
  } // if (!hexmode) ...    printf("// GM voice data in MA-2 format\r\xa\r\xa");
  for (j = 0; j < nvoice; j++) { // read instrument name
    fread(instname[j], 16, 1, stdin);
//    printf("%.3d:%16s\r\xa", j, instname[j]);
  } // for (j
  lvap = &(lva[0]);
  for (i = 0; i < nvoice; i++) {
    k = 0;
    for (j = 0; j < 5; j++) { // copy bank#/inst#/LFO/FB/ALG
      if (cbf[j]) fread(&(vbuf[k++]), 1, 1, stdin); // read 1 byte
      else        fread(dum,          1, 1, stdin); // skip 1 byte
    }
    for (op = 0; op < 2; op++) { // for OP1/OP2
      for (j = 5; j < 10; j++) { // copy OP data
        if (cbf[j]) fread(&(vbuf[k++]), 1, 1, stdin); // read 1 byte
        else        fread(dum,          1, 1, stdin); // skip 1 byte
      } // for (j =
      for (j = 10; j < 14; j++) { // insert zeros for alignment
        if (cbf[j]) vbuf[k++] = 0; // store zero
      } // for (j = 
    } // for (op = 
    for (j = 15; j < DATALEN; j++) {
      fread(dum, 1, 1, stdin); // skip 1 byte
    }
    memcpy(lvap, vbuf, lva_inc); // copy voice data to linear arry
    lvap += lva_inc; // increment pointer
    if (0 == hexmode) {
      if (i) {
        printf(", //%3d:%s\r\xa", i-1, instname[i-1]);
      } // if (i)
      printf("\r\xa    {");
      k = 0;
      for (j = 0; j < 4; j++) {
        if (cbf[j]) {
          if (k) printf(", ");
          printf("0x%.2x", vbuf[k++]);
        }
      } // for (j
      for (j = 5; j < 14 ;j++) {
        if (cbf[j]) {
          if (k) printf(", ");
          printf("0x%.2x", vbuf[k++]); 
        } // if (cbf[j])
      } // for (j
      printf(",\r\xa      %s     ", banner1);
      for (j = 5; j < 13 ;j++) {
        if (cbf[j]) {
          if (5 < j) printf(", ");
          printf("0x%.2x", vbuf[k++]); 
        } // if (cbf[j])
      } // for (j
      printf("}");
    } // if (0 == hexmode) ...
  } // for (i
  i = nvoice;
  if (0 == hexmode) {
  printf("  //%3d:%s", i-1, instname[i-1]);
  printf("\r\xa}; // const ma2vdat vma2[] PROGMEM\r\xa");
  } // if (!hexmode) ...
  else if (1 == hexmode) { // intel hex mode, 
    lvap = &(lva[0]);
    for (i = 0; i < (nvoice*lva_inc+rlen-1) / rlen; i++) {
      csum = 0;
      printf(":");
      outhex(rlen);
      outhex(adrs >> 8);
      outhex(adrs     );
      outhex(0);
      for (j = 0; j < rlen ;j++) {
        outhex(*(lvap++));
      } // for (j
      outhex(0x100 - csum);
      printf("\r\xa");
      adrs += rlen;
    } // for (i= 0; ...
    if (endrecord) {
      csum = 0;
      rlen = 0;
      adrs = 0;
      printf(":");
      outhex(rlen);
      outhex(adrs >> 8);
      outhex(adrs     );
      outhex(1);
      outhex(0x100 - csum);
      printf("\r\xa");
    } // if (endrecord)
  }   // else if (1 == hexmode)
  else if (2 == hexmode) { // Motorola S-format
    lvap = &(lva[0]);
    for (i = 0; i < (nvoice*lva_inc+rlen-1) / rlen; i++) {
      csum = 0;
      printf("S2");
      outhex(rlen+4);
      outhex(adrs >> 16);
      outhex(adrs >> 8);
      outhex(adrs     );
      for (j = 0; j < rlen ;j++) {
        outhex(*(lvap++));
      } // for (j
      outhex(0xFF ^ csum);
      printf("\r\xa");
      adrs += rlen;
    } // for (i= 0; ...
    if (endrecord) {
      csum = 0;
      rlen = 4;
      adrs = 0;
      printf("S8");
      outhex(rlen);
      outhex(adrs >> 16);
      outhex(adrs >> 8);
      outhex(adrs     );
      outhex(0xFF ^ csum);
      printf("\r\xa");
    } // if (endrecord)
  } // else if (2 == hexmode)
  return(0);
} // int main()
