/*************************************************************************/
/*   MIDI tone generator (FM synthesis) with demo song (SMAF sequencer)  */
/*   for CQ publishing CQ-FRK-SH2A board (Renesas SH72620)               */
/*   using HEW 4 environment                                             */
/*                                                                       */
/*   NON COMMERCIAL USE ONLY                                             */
/*                                                                       */
/*   2010/05/12 Rev. 1.01 by pcm1723                                     */
/*                                                                       */
/*************************************************************************/

#include "tonegen.h"

// tone generator main
int tg_main(void)
{
  volatile uint16_t count=0;
#ifndef USEWAVEQUE
  uint32_t lw;
#endif
	long int w0;
#ifdef REVERB
  volatile long w1;
#endif
  int i;
  volatile long int sum;
//  ch_prop_t   *c_p;
#ifdef __v850__
  int ret;
	/* CPU initialization */
  VSWC	= 1;
  WDTM2	= 0;

  __asm("mov	0x00, r11");
  __asm("st.b	r11,  PRCMD");
  __asm("st.b	r11,  PCC");

  while( LOCK );
  SELPLL	= 1;

  ret	= _rcopy( &_S_romp, -1 );
#endif

#ifdef SEQUENCER
  demosave = demoflag;
  demo_no = 0;
  #ifdef LCD
    demoflag = 0; 
  #endif
#endif                              
#ifdef ADUC7026
  arm_port_setup();
#endif
#ifdef __v850__
  v850_port_setup();
#endif
#ifdef STM32F103
  starm_port_setup();
#endif
#ifdef LPC23XX
  LPC23xx_port_setup();
#endif
#ifdef _SH2AFPU
  SH2A_port_setup();
#endif
/*
// search for a configuration with fs >= 15 kHz
  for (  i  = MAX_CONFIG - 1;
       (15 > sys_config[fastchip][i].tick1ms) & (0 < i);
       i--) {} 
//  sys_mode_index = i; 
#ifdef DEMOONLY
  sys_mode_index = 0;
#else
  sys_mode_index = 3;
#endif
*/
  sys_mode_index = 2;
  
  last_note = -1;
  midi_mon_part = 0; // MIDI monitor channel
  wq_wr_ix = 0; // wave queue write index = 0
  wq_rd_ix = 0; // wave queue read  index = 0
// fill up wave queue with 0
  for (i = 0; i < WAVEQUESIZE; i++) {wave_que[i] = 0;}
#ifdef REVERB
  for (i = 0; i < DELAYMEMSIZE; i++) {delaymem[i] = 0;}
  delayp = 0;
  echo_dep = 128;
#endif
  set_config(sys_mode_index);
  hard_reset(); // hard reset
  enable_interrupt();
#ifdef LCD
  panel_mode = MIDI_MON;
  lcd_init();
  lcd_draw_panel(panel_mode);
  lcd_buf_ind = 0;
#else 
  #ifdef SEQUENCER
    init_seq();
  #endif
#endif
  w0  = 0;
  sum = 0; // clear initial wave output
  
  while (1){ // Main sampling rate loop
    if (CHECKMIDI) { // serial received
     uint8_t temp;
     LED_OFF;        // OFF P4.7iLED D1j
     mididec(GETMIDI); // decode MIDI data
//     LED_ON;         // ON P4.7iLED D1j
  } // if (CHECKMIDI) ...
  else
#ifdef USEWAVEQUE
    if(wq_wr_ix != wq_rd_ix) // wave queue has room to write
#else
    if(t1_tick)
#endif
  {
#ifdef REVERB
    sum += (w0 >> 1); // mix dry 50% : reverb 50 %
#endif

#ifdef syn_CV
    switch (0x03 & count) { // SPI xmit timing
      case 0:
        if (CHECK_SPIQ) {
          if (SPI_1ST_BYTE) SS_LOW;  // make SS low before 1st byte
          XMIT_SPI;
        } // if (CHECK_SPIQ)
        break;
      case 2:
        if (SPI_1ST_BYTE) SS_HIGH; // make SS high for idle
        break;
      default:
        break;
    } // switch ()
#endif

#ifdef USEWAVEQUE
//sum = 0xffff & (wq_wr_ix << 6);    
//sum = 0xffff & (seq_p << 6);    
//sum = (count << 10);
    wave_que[wq_wr_ix] = sum;
    wq_wr_ix = ((wq_wr_ix + 1) & (WAVEQUESIZE-1));
#else
		lw = (unsigned long)(0x8000 ^ sum) << 12; // adjust for DAC bit allocation
		DAC3DAT = lw; // output wave to DAC3
#endif
    t1_tick =0;   // clear T1 tick flag
    LED_OFF;       // OFF P4.7iLED D1j
    
#ifdef USEWAVEQUE
    if (1 == tick1ms) { // EG update timing
      for (i = 0; i < MAXUNITS; i++) {
        unit_prop_t *u_p = &(unit_prop[i]);
        eg(u_p, &(u_p->op_prop[0]));
        eg(u_p, &(u_p->op_prop[1]));
      } // for 
    } // if 
#else
    if      (1 == tick1ms) { // EG for unit 0
      unit_prop_t *u_p = &(unit_prop[0]);
      eg(u_p, &(u_p->op_prop[0]));
      eg(u_p, &(u_p->op_prop[1]));
    }
    else if (5 == tick1ms) { // EG for unit 1
      unit_prop_t *u_p = &(unit_prop[1]);
      eg(u_p, &(u_p->op_prop[0]));
      eg(u_p, &(u_p->op_prop[1]));
    }
    else if (9 == tick1ms) { // EG for unit 2
      unit_prop_t *u_p = &(unit_prop[2]);
      eg(u_p, &(u_p->op_prop[0]));
      eg(u_p, &(u_p->op_prop[1]));
    }
#endif
// combine output of 3 units with 1/4 scaling to avoid overflow
    sum= 0;
    for (i = 0; i < MAXUNITS; i++) {
      int32_t ssum = sum;
      sum += unit(&(unit_prop[i])) >> 2; 
//      if (2 != i) sum = ssum;
    } // for

#ifdef REVERB
    w0 = delaymem[delayp]; // read sample from delay memory
    w1 = ((w0+1) >> 1)                              // feedback of delayed sample
       + ((sum * ch_prop[0].MD_rev_send) >> 7); // mult by reverb send level of ch0
// too high send level may cause overflow, apply clipping if so
    if      ( MAXSHORT < w1) w1 =  MAXSHORT;
    else if (-MAXSHORT > w1) w1 = -MAXSHORT;
    delaymem[delayp++] = w1; // write sample in delay memory
    if (delay_mem_size <= delayp) delayp = 0; // modulo operation
#endif

    count++;
    count &= 0x3f;
    if (0 >= --tick1ms) { // 1 ms tick timing
      tick1ms = SEQTICK1ms; // restore 1ms counter
#ifdef LCD
      if (0 > --lcd_blink_timer) { // LCD blink timer
        lcd_blink_timer = LCD_BLINK_ms - 1;
        if (&(lcd_buf[0]) == lcd_buf_p) {
          lcd_blink_flag  ^= 1; // flip blink flag
        } else {                // sub screen displayed
          lcd_blink_flag   = 0; // do not blink if so
        } // else
      } // if (0 > --lcd_blink_timer) 
#endif
      t4ms++;
      t4ms &= 0x3; // divide by 4 with 1ms tick
      if (MAXUNITS > t4ms) {
        update_LFO(&(ch_prop[t4ms])); // update each channel LFO at 4ms interval
        update_porta(&(unit_prop[t4ms])); // update portament of each unit at 4ms
      } // if (MAXUNITS > t4ms)
      if (MAXUNITS > (t4ms+4)) {
        update_LFO(&(ch_prop[t4ms+4])); // update each channel LFO at 4ms interval
        update_porta(&(unit_prop[t4ms+4])); // update portament of each unit at 4ms
      } // if (MAXUNITS > t4ms)
#ifdef LCD      
      if (0 >= --debounce) { // debounce interval
        debounce = DEBOUNCE_ms; // restore debounce counter
        if ((sw_val = sw_input())) {  // INC or DEC switch released
#ifdef SEQUENCER
          if (demoflag | demosave) {     // in demo mode
            switch (abs(sw_val)) { //
              case CLICK :    // skip to next song
                seq_songsel(demo_no+sw_val);
                break;
              case PRESS :    // exit from demo mode
                demo_off();
                break;
              default: break;
            } // switch
          } // if (demoflag)
          else 
#endif
            panel_cmd(sw_val); // other commands
        } // if ((sw_val == sw_input())) ...
        else {
          sw_val = NOCHANGE;
          if (SET_CONFIG == panel_mode) lcd_cpuload();
        } // else
      } // if (0 >= debounce)
#endif      
    } // if (0 >= --tick1ms)
#ifdef LCD
    if (4 == tick1ms) {
      switch (t4ms) { 
        case 3:  // force E trigger off
          lcd_E_down();
          if (lcd_disp_timer) { // LCD display timer working
            if (0 == --lcd_disp_timer) {
              lcd_buf_p = &(lcd_buf[0]); // restore pointer to normal screen
#ifdef SEQUENCER
              if (demosave) {
                demoflag = demosave;
                demosave = 0;
                init_seq();
              } // if (demosave)
#endif
            } // if
          } // if 
          break;
#ifdef LCD_4bit
        case 2: // LCD low nibble output timing
          lcd_refresh(1);
          break;
        case 1: // force E trigger off
          lcd_E_down();
          break;
#endif
        case 0: // LCD high nibble output timing
          lcd_refresh(0);
          break;
        default: break;
      } // switch 
    } // if (4 == tick1ms)
#endif
//    LED_ON;       // ON P4.7iLED D1j
#ifdef SEQUENCER
    if (demoflag) seq(); // SMAF sequencer top level
#endif
    LED_ON;       // ON P4.7iLED D1j
    }  // if 
  } // while (1) 
} // int main()
